// Copyright Varjo Technologies Oy. All Rights Reserved.

#include "VarjoVRController.h"

#include <string>

#define LOCTEXT_NAMESPACE "VarjoVRController"

FVarjoHMD* FVarjoVRController::GetVarjoHMD() const
{
	static FName SystemName(TEXT("VarjoHMD"));
	if (GEngine->XRSystem.IsValid() && (GEngine->XRSystem->GetSystemName() == SystemName)) {
		return static_cast<FVarjoHMD*>(GEngine->XRSystem.Get());
	}

	return nullptr;
}

FVarjoVRController::FVarjoVRController(const TSharedRef< FGenericApplicationMessageHandler >& InMessageHandler)
	: MessageHandler(InMessageHandler),
	VarjoHMDPlugin(nullptr)
{

	FMemory::Memzero(ControllerStates, sizeof(ControllerStates));

	for (int32 i=0; i < vr::k_unMaxTrackedDeviceCount; ++i) {
		DeviceToControllerMap[i] = INDEX_NONE;
	}

	for (int32 UnrealControllerIndex = 0; UnrealControllerIndex < MaxUnrealControllers; ++UnrealControllerIndex) {
		for (int32 HandIndex = 0; HandIndex < vr::k_unMaxTrackedDeviceCount; ++HandIndex) {
			UnrealControllerIdAndHandToDeviceIdMap[UnrealControllerIndex][HandIndex] = INDEX_NONE;
		}
	}

	for (int32& Count : UnrealControllerHandUsageCount) {
		Count = 0;
	}

	NumControllersMapped = 0;
	NumTrackersMapped = 0;

	InitialButtonRepeatDelay = 0.2f;
	ButtonRepeatDelay = 0.1f;

	Buttons[ (int32)EControllerHand::Left ][ EVarjoVRControllerButton::System ] = FGamepadKeyNames::SpecialLeft;
	Buttons[ (int32)EControllerHand::Left ][ EVarjoVRControllerButton::ApplicationMenu ] = FGamepadKeyNames::MotionController_Left_Shoulder;
	Buttons[ (int32)EControllerHand::Left ][ EVarjoVRControllerButton::TouchPadPress ] = FGamepadKeyNames::MotionController_Left_Thumbstick;
	Buttons[ (int32)EControllerHand::Left ][ EVarjoVRControllerButton::TouchPadTouch ] = VarjoVRControllerKeyNames::Touch0;
	Buttons[ (int32)EControllerHand::Left ][ EVarjoVRControllerButton::TriggerPress ] = FGamepadKeyNames::MotionController_Left_Trigger;
	Buttons[ (int32)EControllerHand::Left ][ EVarjoVRControllerButton::Grip ] = FGamepadKeyNames::MotionController_Left_Grip1;

	Buttons[ (int32)EControllerHand::Right ][ EVarjoVRControllerButton::System ] = FGamepadKeyNames::SpecialRight;
	Buttons[ (int32)EControllerHand::Right ][ EVarjoVRControllerButton::ApplicationMenu ] = FGamepadKeyNames::MotionController_Right_Shoulder;
	Buttons[ (int32)EControllerHand::Right ][ EVarjoVRControllerButton::TouchPadPress ] = FGamepadKeyNames::MotionController_Right_Thumbstick;
	Buttons[ (int32)EControllerHand::Right ][ EVarjoVRControllerButton::TouchPadTouch ] = VarjoVRControllerKeyNames::Touch1;
	Buttons[ (int32)EControllerHand::Right ][ EVarjoVRControllerButton::TriggerPress ] = FGamepadKeyNames::MotionController_Right_Trigger;
	Buttons[ (int32)EControllerHand::Right ][ EVarjoVRControllerButton::Grip ] = FGamepadKeyNames::MotionController_Right_Grip1;

	// Init Left & Right, TouchPadUp/Down/Left/Right button mappings
	SetTouchDPadMapping(DefaultDPadMapping);

	Buttons[ (int32)EControllerHand::Pad ][ EVarjoVRControllerButton::System ] = VarjoVRControllerKeyNames::GenericSystem;
	Buttons[ (int32)EControllerHand::Pad ][ EVarjoVRControllerButton::ApplicationMenu ] = VarjoVRControllerKeyNames::GenericMenu;
	Buttons[ (int32)EControllerHand::Pad ][ EVarjoVRControllerButton::TouchPadPress ] = VarjoVRControllerKeyNames::GenericTouchpad;
	Buttons[ (int32)EControllerHand::Pad ][ EVarjoVRControllerButton::TouchPadTouch ] = FGamepadKeyNames::Invalid;
	Buttons[ (int32)EControllerHand::Pad ][ EVarjoVRControllerButton::TriggerPress ] = VarjoVRControllerKeyNames::GenericTrigger;
	Buttons[ (int32)EControllerHand::Pad ][ EVarjoVRControllerButton::Grip ] = VarjoVRControllerKeyNames::GenericGrip;
	Buttons[ (int32)EControllerHand::Pad ][ EVarjoVRControllerButton::TouchPadUp ] = FGamepadKeyNames::Invalid;
	Buttons[ (int32)EControllerHand::Pad ][ EVarjoVRControllerButton::TouchPadDown ] = FGamepadKeyNames::Invalid;
	Buttons[ (int32)EControllerHand::Pad ][ EVarjoVRControllerButton::TouchPadLeft ] = FGamepadKeyNames::Invalid;
	Buttons[ (int32)EControllerHand::Pad ][ EVarjoVRControllerButton::TouchPadRight ] = FGamepadKeyNames::Invalid;

	Buttons[ (int32)EControllerHand::ExternalCamera ][ EVarjoVRControllerButton::System ] = FGamepadKeyNames::Invalid;
	Buttons[ (int32)EControllerHand::ExternalCamera ][ EVarjoVRControllerButton::ApplicationMenu ] = FGamepadKeyNames::Invalid;
	Buttons[ (int32)EControllerHand::ExternalCamera ][ EVarjoVRControllerButton::TouchPadPress ] = FGamepadKeyNames::Invalid;
	Buttons[ (int32)EControllerHand::ExternalCamera ][ EVarjoVRControllerButton::TouchPadTouch ] = FGamepadKeyNames::Invalid;
	Buttons[ (int32)EControllerHand::ExternalCamera ][ EVarjoVRControllerButton::TriggerPress ] = FGamepadKeyNames::Invalid;
	Buttons[ (int32)EControllerHand::ExternalCamera ][ EVarjoVRControllerButton::Grip ] = FGamepadKeyNames::Invalid;
	Buttons[ (int32)EControllerHand::ExternalCamera ][ EVarjoVRControllerButton::TouchPadUp ] = FGamepadKeyNames::Invalid;
	Buttons[ (int32)EControllerHand::ExternalCamera ][ EVarjoVRControllerButton::TouchPadDown ] = FGamepadKeyNames::Invalid;
	Buttons[ (int32)EControllerHand::ExternalCamera ][ EVarjoVRControllerButton::TouchPadLeft ] = FGamepadKeyNames::Invalid;
	Buttons[ (int32)EControllerHand::ExternalCamera ][ EVarjoVRControllerButton::TouchPadRight ] = FGamepadKeyNames::Invalid;

	Buttons[(int32)EControllerHand::Gun][EVarjoVRControllerButton::System] = VarjoVRControllerKeyNames::GenericSystem;
	Buttons[(int32)EControllerHand::Gun][EVarjoVRControllerButton::ApplicationMenu] = VarjoVRControllerKeyNames::GenericMenu;
	Buttons[(int32)EControllerHand::Gun][EVarjoVRControllerButton::TouchPadPress] = FGamepadKeyNames::Invalid;
	Buttons[(int32)EControllerHand::Gun][EVarjoVRControllerButton::TouchPadTouch] = FGamepadKeyNames::Invalid;
	Buttons[(int32)EControllerHand::Gun][EVarjoVRControllerButton::TriggerPress] = VarjoVRControllerKeyNames::GenericTrigger;
	Buttons[(int32)EControllerHand::Gun][EVarjoVRControllerButton::Grip] = VarjoVRControllerKeyNames::GenericGrip;
	Buttons[(int32)EControllerHand::Gun][EVarjoVRControllerButton::TouchPadUp] = FGamepadKeyNames::Invalid;
	Buttons[(int32)EControllerHand::Gun][EVarjoVRControllerButton::TouchPadDown] = FGamepadKeyNames::Invalid;
	Buttons[(int32)EControllerHand::Gun][EVarjoVRControllerButton::TouchPadLeft] = FGamepadKeyNames::Invalid;
	Buttons[(int32)EControllerHand::Gun][EVarjoVRControllerButton::TouchPadRight] = FGamepadKeyNames::Invalid;

	for (int32 SpecialIndex = (int32)EControllerHand::Special_1; SpecialIndex <= (int32)EControllerHand::Special_9; ++SpecialIndex) {
		Buttons[SpecialIndex][EVarjoVRControllerButton::System] = VarjoVRControllerKeyNames::GenericSystem;
		Buttons[SpecialIndex][EVarjoVRControllerButton::ApplicationMenu] = VarjoVRControllerKeyNames::GenericMenu;
		Buttons[SpecialIndex][EVarjoVRControllerButton::TouchPadPress] = VarjoVRControllerKeyNames::GenericTouchpad;
		Buttons[SpecialIndex][EVarjoVRControllerButton::TouchPadTouch] = FGamepadKeyNames::Invalid;
		Buttons[SpecialIndex][EVarjoVRControllerButton::TriggerPress] = VarjoVRControllerKeyNames::GenericTrigger;
		Buttons[SpecialIndex][EVarjoVRControllerButton::Grip] = VarjoVRControllerKeyNames::GenericGrip;
		Buttons[SpecialIndex][EVarjoVRControllerButton::TouchPadUp] = FGamepadKeyNames::Invalid;
		Buttons[SpecialIndex][EVarjoVRControllerButton::TouchPadDown] = FGamepadKeyNames::Invalid;
		Buttons[SpecialIndex][EVarjoVRControllerButton::TouchPadLeft] = FGamepadKeyNames::Invalid;
		Buttons[SpecialIndex][EVarjoVRControllerButton::TouchPadRight] = FGamepadKeyNames::Invalid;
	}

	EKeys::AddKey(FKeyDetails(VarjoVRControllerKeys::VarjoVR_Knuckles_Left_HandGrip, LOCTEXT("VarjoVR_Knuckles_Left_HandGrip", "VarjoVR Knuckles (L) Hand Grip CapSense"), FKeyDetails::GamepadKey | FKeyDetails::FloatAxis));
	EKeys::AddKey(FKeyDetails(VarjoVRControllerKeys::VarjoVR_Knuckles_Left_IndexGrip, LOCTEXT("VarjoVR_Knuckles_Left_IndexGrip", "VarjoVR Knuckles (L) Index Grip CapSense"), FKeyDetails::GamepadKey | FKeyDetails::FloatAxis));
	EKeys::AddKey(FKeyDetails(VarjoVRControllerKeys::VarjoVR_Knuckles_Left_MiddleGrip, LOCTEXT("VarjoVR_Knuckles_Left_MiddleGrip", "VarjoVR Knuckles (L) Middle Grip CapSense"), FKeyDetails::GamepadKey | FKeyDetails::FloatAxis));
	EKeys::AddKey(FKeyDetails(VarjoVRControllerKeys::VarjoVR_Knuckles_Left_RingGrip, LOCTEXT("VarjoVR_Knuckles_Left_RingGrip", "VarjoVR Knuckles (L) Ring Grip CapSense"), FKeyDetails::GamepadKey | FKeyDetails::FloatAxis));
	EKeys::AddKey(FKeyDetails(VarjoVRControllerKeys::VarjoVR_Knuckles_Left_PinkyGrip, LOCTEXT("VarjoVR_Knuckles_Left_PinkyGrip", "VarjoVR Knuckles (L) Pinky Grip CapSense"), FKeyDetails::GamepadKey | FKeyDetails::FloatAxis));

	EKeys::AddKey(FKeyDetails(VarjoVRControllerKeys::VarjoVR_Knuckles_Right_HandGrip, LOCTEXT("VarjoVR_Knuckles_Right_HandGrip", "VarjoVR Knuckles (R) Hand Grip CapSense"), FKeyDetails::GamepadKey | FKeyDetails::FloatAxis));
	EKeys::AddKey(FKeyDetails(VarjoVRControllerKeys::VarjoVR_Knuckles_Right_IndexGrip, LOCTEXT("VarjoVR_Knuckles_Right_IndexGrip", "VarjoVR Knuckles (R) Index Grip CapSense"), FKeyDetails::GamepadKey | FKeyDetails::FloatAxis));
	EKeys::AddKey(FKeyDetails(VarjoVRControllerKeys::VarjoVR_Knuckles_Right_MiddleGrip, LOCTEXT("VarjoVR_Knuckles_Right_MiddleGrip", "VarjoVR Knuckles (R) Middle Grip CapSense"), FKeyDetails::GamepadKey | FKeyDetails::FloatAxis));
	EKeys::AddKey(FKeyDetails(VarjoVRControllerKeys::VarjoVR_Knuckles_Right_RingGrip, LOCTEXT("VarjoVR_Knuckles_Right_RingGrip", "VarjoVR Knuckles (R) Ring Grip CapSense"), FKeyDetails::GamepadKey | FKeyDetails::FloatAxis));
	EKeys::AddKey(FKeyDetails(VarjoVRControllerKeys::VarjoVR_Knuckles_Right_PinkyGrip, LOCTEXT("VarjoVR_Knuckles_Right_PinkyGrip", "VarjoVR Knuckles (R) Pinky Grip CapSense"), FKeyDetails::GamepadKey | FKeyDetails::FloatAxis));

	IModularFeatures::Get().RegisterModularFeature(GetModularFeatureName(), this);

}

FVarjoVRController::~FVarjoVRController()
{
	IModularFeatures::Get().UnregisterModularFeature(GetModularFeatureName(), this);
}

void FVarjoVRController::Tick( float DeltaTime )
{
	vr::IVRSystem* VRSystem = GetVRSystem();

	if (VRSystem != nullptr) {
		RegisterDeviceChanges(VRSystem);
		DetectHandednessSwap(VRSystem);
	}
}

void FVarjoVRController::SendControllerEvents()
{
	vr::VRControllerState_t VRControllerState;

	vr::IVRSystem* VRSystem = GetVRSystem();

	if (VRSystem != nullptr) {
		const double CurrentTime = FPlatformTime::Seconds();

		for (uint32 DeviceIndex = 0; DeviceIndex < vr::k_unMaxTrackedDeviceCount; ++DeviceIndex) {
			// see what kind of hardware this is
			vr::ETrackedDeviceClass DeviceClass = VRSystem->GetTrackedDeviceClass(DeviceIndex);

			// skip non-controller or non-tracker devices
			if (DeviceClass != vr::TrackedDeviceClass_Controller && DeviceClass != vr::TrackedDeviceClass_GenericTracker) {
				continue;
			}

			// get the controller index for this device
			int32 ControllerIndex = DeviceToControllerMap[DeviceIndex];
			FControllerState& ControllerState = ControllerStates[ DeviceIndex ];
			EControllerHand HandToUse = ControllerState.Hand;

			// see if this is a hand specific controller
			if (HandToUse == EControllerHand::Left || HandToUse == EControllerHand::Right) {
				// check to see if we need to swap input hands for debugging
				static const auto CVar = IConsoleManager::Get().FindTConsoleVariableDataInt(TEXT("vr.SwapMotionControllerInput"));
				bool bSwapHandInput = (CVar->GetValueOnGameThread() != 0) ? true : false;
				if (bSwapHandInput) {
					HandToUse = (HandToUse == EControllerHand::Left) ? EControllerHand::Right : EControllerHand::Left;
				}
			}
			if (VRSystem->GetControllerState(DeviceIndex, &VRControllerState, sizeof(vr::VRControllerState_t))) {
				if (VRControllerState.unPacketNum != ControllerState.PacketNum ) {
					bool CurrentStates[ EVarjoVRControllerButton::TotalButtonCount ] = {0};

					// Get the current state of all buttons
					CurrentStates[ EVarjoVRControllerButton::System ] = !!(VRControllerState.ulButtonPressed & vr::ButtonMaskFromId(vr::k_EButton_System));
					CurrentStates[ EVarjoVRControllerButton::ApplicationMenu ] = !!(VRControllerState.ulButtonPressed & vr::ButtonMaskFromId(vr::k_EButton_ApplicationMenu));
					CurrentStates[ EVarjoVRControllerButton::TouchPadPress ] = !!(VRControllerState.ulButtonPressed & vr::ButtonMaskFromId(vr::k_EButton_SteamVR_Touchpad));
					CurrentStates[ EVarjoVRControllerButton::TouchPadTouch ] = !!( VRControllerState.ulButtonTouched & vr::ButtonMaskFromId( vr::k_EButton_SteamVR_Touchpad ) );
					CurrentStates[ EVarjoVRControllerButton::TriggerPress ] = !!(VRControllerState.ulButtonPressed & vr::ButtonMaskFromId(vr::k_EButton_SteamVR_Trigger));
					CurrentStates[ EVarjoVRControllerButton::Grip ] = !!(VRControllerState.ulButtonPressed & vr::ButtonMaskFromId(vr::k_EButton_Grip));

					// If the touchpad isn't currently pressed or touched, zero put both of the axes
					if (!CurrentStates[ EVarjoVRControllerButton::TouchPadTouch ]) {
 						VRControllerState.rAxis[TOUCHPAD_AXIS].y = 0.0f;
 						VRControllerState.rAxis[TOUCHPAD_AXIS].x = 0.0f;
					}

					// D-pad emulation
					const FVector2D TouchDir = FVector2D(VRControllerState.rAxis[TOUCHPAD_AXIS].x, VRControllerState.rAxis[TOUCHPAD_AXIS].y).GetSafeNormal();
					const FVector2D UpDir(0.f, 1.f);
					const FVector2D RightDir(1.f, 0.f);

					const float VerticalDot = TouchDir | UpDir;
					const float RightDot = TouchDir | RightDir;

					const bool bPressed = !TouchDir.IsNearlyZero() && CurrentStates[ EVarjoVRControllerButton::TouchPadPress ];
						
					CurrentStates[ EVarjoVRControllerButton::TouchPadUp ]		= bPressed && (VerticalDot >= DOT_45DEG);
					CurrentStates[ EVarjoVRControllerButton::TouchPadDown ]		= bPressed && (VerticalDot <= -DOT_45DEG);
					CurrentStates[ EVarjoVRControllerButton::TouchPadLeft ]		= bPressed && (RightDot <= -DOT_45DEG);
					CurrentStates[ EVarjoVRControllerButton::TouchPadRight ]	= bPressed && (RightDot >= DOT_45DEG);

					if ( ControllerState.TouchPadXAnalog != VRControllerState.rAxis[TOUCHPAD_AXIS].x) {
						const FGamepadKeyNames::Type AxisButton = (HandToUse == EControllerHand::Left) ? FGamepadKeyNames::MotionController_Left_Thumbstick_X : FGamepadKeyNames::MotionController_Right_Thumbstick_X;
						MessageHandler->OnControllerAnalog(AxisButton, ControllerIndex, VRControllerState.rAxis[TOUCHPAD_AXIS].x);
						ControllerState.TouchPadXAnalog = VRControllerState.rAxis[TOUCHPAD_AXIS].x;
					}

					if ( ControllerState.TouchPadYAnalog != VRControllerState.rAxis[TOUCHPAD_AXIS].y) {
						const FGamepadKeyNames::Type AxisButton = (HandToUse == EControllerHand::Left) ? FGamepadKeyNames::MotionController_Left_Thumbstick_Y : FGamepadKeyNames::MotionController_Right_Thumbstick_Y;
						// Invert the y to match UE4 convention
						const float Value = -VRControllerState.rAxis[TOUCHPAD_AXIS].y;
						MessageHandler->OnControllerAnalog(AxisButton, ControllerIndex, Value);
						ControllerState.TouchPadYAnalog = Value;
					}

					if ( ControllerState.TriggerAnalog != VRControllerState.rAxis[TRIGGER_AXIS].x) {
						const FGamepadKeyNames::Type AxisButton = (HandToUse == EControllerHand::Left) ? FGamepadKeyNames::MotionController_Left_TriggerAxis : FGamepadKeyNames::MotionController_Right_TriggerAxis;
						MessageHandler->OnControllerAnalog(AxisButton, ControllerIndex, VRControllerState.rAxis[TRIGGER_AXIS].x);
						ControllerState.TriggerAnalog = VRControllerState.rAxis[TRIGGER_AXIS].x;
					}

					// Knuckles CapSense Grip Axes Updates
					{
						if (ControllerState.HandGripAnalog != VRControllerState.rAxis[KNUCKLES_TOTAL_HAND_GRIP_AXIS].x) {
							const FGamepadKeyNames::Type AxisButton = (HandToUse == EControllerHand::Left) ? VarjoVRControllerKeyNames::VarjoVR_Knuckles_Left_HandGrip : VarjoVRControllerKeyNames::VarjoVR_Knuckles_Right_HandGrip;
							MessageHandler->OnControllerAnalog(AxisButton, ControllerIndex, VRControllerState.rAxis[KNUCKLES_TOTAL_HAND_GRIP_AXIS].x);
							ControllerState.HandGripAnalog = VRControllerState.rAxis[KNUCKLES_TOTAL_HAND_GRIP_AXIS].x;
						}

						if (ControllerState.IndexGripAnalog != VRControllerState.rAxis[KNUCKLES_UPPER_HAND_GRIP_AXIS].x) {
							const FGamepadKeyNames::Type AxisButton = (HandToUse == EControllerHand::Left) ? VarjoVRControllerKeyNames::VarjoVR_Knuckles_Left_IndexGrip : VarjoVRControllerKeyNames::VarjoVR_Knuckles_Right_IndexGrip;
							MessageHandler->OnControllerAnalog(AxisButton, ControllerIndex, VRControllerState.rAxis[KNUCKLES_UPPER_HAND_GRIP_AXIS].x);
							ControllerState.IndexGripAnalog = VRControllerState.rAxis[KNUCKLES_UPPER_HAND_GRIP_AXIS].x;
						}

						if (ControllerState.MiddleGripAnalog != VRControllerState.rAxis[KNUCKLES_UPPER_HAND_GRIP_AXIS].y) {
							const FGamepadKeyNames::Type AxisButton = (HandToUse == EControllerHand::Left) ? VarjoVRControllerKeyNames::VarjoVR_Knuckles_Left_MiddleGrip : VarjoVRControllerKeyNames::VarjoVR_Knuckles_Right_MiddleGrip;
							MessageHandler->OnControllerAnalog(AxisButton, ControllerIndex, VRControllerState.rAxis[KNUCKLES_UPPER_HAND_GRIP_AXIS].y);
							ControllerState.MiddleGripAnalog = VRControllerState.rAxis[KNUCKLES_UPPER_HAND_GRIP_AXIS].y;
						}

						if (ControllerState.RingGripAnalog != VRControllerState.rAxis[KNUCKLES_LOWER_HAND_GRIP_AXIS].x) {
							const FGamepadKeyNames::Type AxisButton = (HandToUse == EControllerHand::Left) ? VarjoVRControllerKeyNames::VarjoVR_Knuckles_Left_RingGrip : VarjoVRControllerKeyNames::VarjoVR_Knuckles_Right_RingGrip;
							MessageHandler->OnControllerAnalog(AxisButton, ControllerIndex, VRControllerState.rAxis[KNUCKLES_LOWER_HAND_GRIP_AXIS].x);
							ControllerState.RingGripAnalog = VRControllerState.rAxis[KNUCKLES_LOWER_HAND_GRIP_AXIS].x;
						}

						if (ControllerState.PinkyGripAnalog != VRControllerState.rAxis[KNUCKLES_LOWER_HAND_GRIP_AXIS].y) {
							const FGamepadKeyNames::Type AxisButton = (HandToUse == EControllerHand::Left) ? VarjoVRControllerKeyNames::VarjoVR_Knuckles_Left_PinkyGrip : VarjoVRControllerKeyNames::VarjoVR_Knuckles_Right_PinkyGrip;
							MessageHandler->OnControllerAnalog(AxisButton, ControllerIndex, VRControllerState.rAxis[KNUCKLES_LOWER_HAND_GRIP_AXIS].y);
							ControllerState.PinkyGripAnalog = VRControllerState.rAxis[KNUCKLES_LOWER_HAND_GRIP_AXIS].y;
						}
					}

					// For each button check against the previous state and send the correct message if any
					for (int32 ButtonIndex = 0; ButtonIndex < EVarjoVRControllerButton::TotalButtonCount; ++ButtonIndex) {
						if (CurrentStates[ButtonIndex] != ControllerState.ButtonStates[ButtonIndex]) {
							const FGamepadKeyNames::Type ButtonId = Buttons[(uint8)HandToUse][ButtonIndex];
							if (ButtonId != FGamepadKeyNames::Invalid) {
								if (CurrentStates[ButtonIndex]) {
									MessageHandler->OnControllerButtonPressed(ButtonId, ControllerIndex, /*IsRepeat =*/false);
								}
								else {
									MessageHandler->OnControllerButtonReleased(ButtonId, ControllerIndex, /*IsRepeat =*/false);
								}
							}

							if (CurrentStates[ButtonIndex] != 0) {
								// this button was pressed - set the button's NextRepeatTime to the InitialButtonRepeatDelay
								ControllerState.NextRepeatTime[ButtonIndex] = CurrentTime + InitialButtonRepeatDelay;
							}
						}

						// Update the state for next time
						ControllerState.ButtonStates[ButtonIndex] = CurrentStates[ButtonIndex];
					}

					ControllerState.PacketNum = VRControllerState.unPacketNum;
				}
			}

			for (int32 ButtonIndex = 0; ButtonIndex < EVarjoVRControllerButton::TotalButtonCount; ++ButtonIndex) {
				if ( ControllerState.ButtonStates[ButtonIndex] != 0 && ControllerState.NextRepeatTime[ButtonIndex] <= CurrentTime) {
					const FGamepadKeyNames::Type ButtonId = Buttons[(uint8)HandToUse][ButtonIndex];
					if (ButtonId != FGamepadKeyNames::Invalid) {
						MessageHandler->OnControllerButtonPressed(ButtonId, ControllerIndex, /*IsRepeat =*/true);
					}

					// set the button's NextRepeatTime to the ButtonRepeatDelay
					ControllerState.NextRepeatTime[ButtonIndex] = CurrentTime + ButtonRepeatDelay;
				}
			}
		}
	}
}

void FVarjoVRController::SetTouchDPadMapping(EVarjoVRTouchDPadMapping NewMapping)
{
	DefaultDPadMapping = NewMapping;

	switch (NewMapping)
	{
	case EVarjoVRTouchDPadMapping::FaceButtons:
		Buttons[(int32)EControllerHand::Left][EVarjoVRControllerButton::TouchPadUp]     = FGamepadKeyNames::MotionController_Left_FaceButton1;
		Buttons[(int32)EControllerHand::Left][EVarjoVRControllerButton::TouchPadDown]   = FGamepadKeyNames::MotionController_Left_FaceButton3;
		Buttons[(int32)EControllerHand::Left][EVarjoVRControllerButton::TouchPadLeft]   = FGamepadKeyNames::MotionController_Left_FaceButton4;
		Buttons[(int32)EControllerHand::Left][EVarjoVRControllerButton::TouchPadRight]  = FGamepadKeyNames::MotionController_Left_FaceButton2;

		Buttons[(int32)EControllerHand::Right][EVarjoVRControllerButton::TouchPadUp]    = FGamepadKeyNames::MotionController_Right_FaceButton1;
		Buttons[(int32)EControllerHand::Right][EVarjoVRControllerButton::TouchPadDown]  = FGamepadKeyNames::MotionController_Right_FaceButton3;
		Buttons[(int32)EControllerHand::Right][EVarjoVRControllerButton::TouchPadLeft]  = FGamepadKeyNames::MotionController_Right_FaceButton4;
		Buttons[(int32)EControllerHand::Right][EVarjoVRControllerButton::TouchPadRight] = FGamepadKeyNames::MotionController_Right_FaceButton2;
		break;

	case EVarjoVRTouchDPadMapping::ThumbstickDirections:
		Buttons[(int32)EControllerHand::Left][EVarjoVRControllerButton::TouchPadUp]     = FGamepadKeyNames::MotionController_Left_Thumbstick_Up;
		Buttons[(int32)EControllerHand::Left][EVarjoVRControllerButton::TouchPadDown]   = FGamepadKeyNames::MotionController_Left_Thumbstick_Down;
		Buttons[(int32)EControllerHand::Left][EVarjoVRControllerButton::TouchPadLeft]   = FGamepadKeyNames::MotionController_Left_Thumbstick_Left;
		Buttons[(int32)EControllerHand::Left][EVarjoVRControllerButton::TouchPadRight]  = FGamepadKeyNames::MotionController_Left_Thumbstick_Right;

		Buttons[(int32)EControllerHand::Right][EVarjoVRControllerButton::TouchPadUp]    = FGamepadKeyNames::MotionController_Right_Thumbstick_Up;
		Buttons[(int32)EControllerHand::Right][EVarjoVRControllerButton::TouchPadDown]  = FGamepadKeyNames::MotionController_Right_Thumbstick_Down;
		Buttons[(int32)EControllerHand::Right][EVarjoVRControllerButton::TouchPadLeft]  = FGamepadKeyNames::MotionController_Right_Thumbstick_Left;
		Buttons[(int32)EControllerHand::Right][EVarjoVRControllerButton::TouchPadRight] = FGamepadKeyNames::MotionController_Right_Thumbstick_Right;
		break;

	default:
		UE_LOG(LogVarjoVRController, Warning, TEXT("Unsupported d-pad mapping (%d). Defaulting to disabled."), (int32)NewMapping);
	case EVarjoVRTouchDPadMapping::Disabled:
		Buttons[(int32)EControllerHand::Left][EVarjoVRControllerButton::TouchPadUp]     = FGamepadKeyNames::Invalid;
		Buttons[(int32)EControllerHand::Left][EVarjoVRControllerButton::TouchPadDown]   = FGamepadKeyNames::Invalid;
		Buttons[(int32)EControllerHand::Left][EVarjoVRControllerButton::TouchPadLeft]   = FGamepadKeyNames::Invalid;
		Buttons[(int32)EControllerHand::Left][EVarjoVRControllerButton::TouchPadRight]  = FGamepadKeyNames::Invalid;

		Buttons[(int32)EControllerHand::Right][EVarjoVRControllerButton::TouchPadUp]    = FGamepadKeyNames::Invalid;
		Buttons[(int32)EControllerHand::Right][EVarjoVRControllerButton::TouchPadDown]  = FGamepadKeyNames::Invalid;
		Buttons[(int32)EControllerHand::Right][EVarjoVRControllerButton::TouchPadLeft]  = FGamepadKeyNames::Invalid;
		Buttons[(int32)EControllerHand::Right][EVarjoVRControllerButton::TouchPadRight] = FGamepadKeyNames::Invalid;
		break;
	}
}

int32 FVarjoVRController::UnrealControllerIdToControllerIndex( const int32 UnrealControllerId, const EControllerHand Hand ) const
{
	return UnrealControllerIdAndHandToDeviceIdMap[UnrealControllerId][(int32)Hand];
}

void FVarjoVRController::SetChannelValue(int32 UnrealControllerId, FForceFeedbackChannelType ChannelType, float Value)
{

	// Skip unless this is the left or right large channel, which we consider to be the only VarjoVRController feedback channel
	if( ChannelType != FForceFeedbackChannelType::LEFT_LARGE && ChannelType != FForceFeedbackChannelType::RIGHT_LARGE ) {
		return;
	}

	const EControllerHand Hand = ( ChannelType == FForceFeedbackChannelType::LEFT_LARGE ) ? EControllerHand::Left : EControllerHand::Right;
	const int32 ControllerIndex = UnrealControllerIdToControllerIndex( UnrealControllerId, Hand );

	if ((ControllerIndex >= 0) && ( ControllerIndex < MaxControllers)) {
		FControllerState& ControllerState = ControllerStates[ ControllerIndex ];

		ControllerState.ForceFeedbackValue = Value;

		UpdateVibration( ControllerIndex );
	}

}

void FVarjoVRController::SetChannelValues(int32 UnrealControllerId, const FForceFeedbackValues& Values)
{

	const int32 LeftControllerIndex = UnrealControllerIdToControllerIndex( UnrealControllerId, EControllerHand::Left );
	if ((LeftControllerIndex >= 0) && ( LeftControllerIndex < MaxControllers)) {
		FControllerState& ControllerState = ControllerStates[ LeftControllerIndex ];
		ControllerState.ForceFeedbackValue = Values.LeftLarge;

		UpdateVibration( LeftControllerIndex );
	}

	const int32 RightControllerIndex = UnrealControllerIdToControllerIndex( UnrealControllerId, EControllerHand::Right );
	if( ( RightControllerIndex >= 0 ) && ( RightControllerIndex < MaxControllers ) ) {
		FControllerState& ControllerState = ControllerStates[ RightControllerIndex ];
		ControllerState.ForceFeedbackValue = Values.RightLarge;

		UpdateVibration( RightControllerIndex );
	}

}

IHapticDevice* FVarjoVRController::GetHapticDevice()
{
	return this;
}	

void FVarjoVRController::SetHapticFeedbackValues(int32 UnrealControllerId, int32 Hand, const FHapticFeedbackValues& Values)
{

	if (Hand != (int32)EControllerHand::Left && Hand != (int32)EControllerHand::Right) {
		return;
	}

	const int32 ControllerIndex = UnrealControllerIdToControllerIndex(UnrealControllerId, (EControllerHand)Hand);
	if (ControllerIndex >= 0 && ControllerIndex < MaxControllers) {
		FControllerState& ControllerState = ControllerStates[ControllerIndex];
		ControllerState.ForceFeedbackValue = (Values.Frequency > 0.0f) ? Values.Amplitude : 0.0f;

		UpdateVibration(ControllerIndex);
	}

}

void FVarjoVRController::GetHapticFrequencyRange(float& MinFrequency, float& MaxFrequency) const
{
	MinFrequency = 0.0f;
	MaxFrequency = 1.0f;
}
	
float FVarjoVRController::GetHapticAmplitudeScale() const
{
	return 1.0f;
}

void FVarjoVRController::UpdateVibration( const int32 ControllerIndex )
{
	const FControllerState& ControllerState = ControllerStates[ ControllerIndex ];
	vr::IVRSystem* VRSystem = GetVRSystem();

	if (VRSystem == nullptr) {
		return;
	}

	// Map the float values from [0,1] to be more reasonable values for the VarjoController.  The docs say that [100,2000] are reasonable values
 	const float LeftIntensity = FMath::Clamp(ControllerState.ForceFeedbackValue * 2000.f, 0.f, 2000.f);
	if (LeftIntensity > 0.f) {
		VRSystem->TriggerHapticPulse(ControllerIndex, TOUCHPAD_AXIS, LeftIntensity);
	}
}

void FVarjoVRController::SetMessageHandler( const TSharedRef< FGenericApplicationMessageHandler >& InMessageHandler )
{
	MessageHandler = InMessageHandler;
}

bool FVarjoVRController::Exec( UWorld* InWorld, const TCHAR* Cmd, FOutputDevice& Ar )
{
	return false;
}

FName FVarjoVRController::GetMotionControllerDeviceTypeName() const
{
	return DeviceTypeName;
}

bool FVarjoVRController::GetControllerOrientationAndPosition(const int32 ControllerIndex, const EControllerHand DeviceHand, FRotator& OutOrientation, FVector& OutPosition, float WorldToMetersScale) const
{
	bool RetVal = false;

 	FVarjoHMD* VarjoHMD = GetVarjoHMD();
 	if (VarjoHMD) {
		int32 DeviceId = UnrealControllerIdAndHandToDeviceIdMap[ControllerIndex][(int32)DeviceHand];
 		FQuat DeviceOrientation = FQuat::Identity;
		// Varjo handles WorldToMetersScale when it reads the controller posrot, so we do not need to use it again here.  Debugging found that they are the same.
		RetVal = VarjoHMD->GetCurrentPose(DeviceId, DeviceOrientation, OutPosition);
 		OutOrientation = DeviceOrientation.Rotator();
 	}

	return RetVal;
}

bool FVarjoVRController::IsControllerConnected(const int32 ControllerIndex, const EControllerHand DeviceHand) const
{
	bool RetVal = false;

	FVarjoHMD* VarjoHMD = GetVarjoHMD();
	if (VarjoHMD) {
		int32 DeviceId = UnrealControllerIdAndHandToDeviceIdMap[ControllerIndex][(int32)DeviceHand];
		FQuat DeviceOrientation = FQuat::Identity;
		RetVal = VarjoHMD->IsDeviceConnected(DeviceId);
	}

	return RetVal;
}

float FVarjoVRController::GetControllerBatteryLevel(const int32 ControllerIndex, const EControllerHand DeviceHand) const
{
	float RetVal = 0.0f;

	FVarjoHMD* VarjoHMD = GetVarjoHMD();
	if (VarjoHMD) {
		int32 DeviceId = UnrealControllerIdAndHandToDeviceIdMap[ControllerIndex][(int32)DeviceHand];
		FQuat DeviceOrientation = FQuat::Identity;
		RetVal = VarjoHMD->GetDeviceBatteryLevel(DeviceId);
	}

	return RetVal;
}

ETrackingStatus FVarjoVRController::GetControllerTrackingStatus(const int32 ControllerIndex, const EControllerHand DeviceHand) const
{
	ETrackingStatus TrackingStatus = ETrackingStatus::NotTracked;

	FVarjoHMD* VarjoHMD = GetVarjoHMD();
 	if (VarjoHMD) {
		int32 DeviceId = UnrealControllerIdAndHandToDeviceIdMap[ControllerIndex][(int32)DeviceHand];
		TrackingStatus = VarjoHMD->GetControllerTrackingStatus(DeviceId);
 	}

	return TrackingStatus;
}

bool FVarjoVRController::IsGamepadAttached() const
{
	FVarjoHMD* VarjoVRSystem = GetVarjoHMD();

	if (VarjoVRSystem != nullptr) {
		// Check if at least one motion controller is tracked
		// Only need to check for at least one player (player index 0)
		int32 PlayerIndex = 0;
		ETrackingStatus LeftHandTrackingStatus = GetControllerTrackingStatus(PlayerIndex, EControllerHand::Left);
		ETrackingStatus RightHandTrackingStatus = GetControllerTrackingStatus(PlayerIndex, EControllerHand::Right);

		return LeftHandTrackingStatus == ETrackingStatus::Tracked || RightHandTrackingStatus == ETrackingStatus::Tracked;
	}

	return false;
}

inline vr::IVRSystem* FVarjoVRController::GetVRSystem()
{
	if (VarjoHMDPlugin == nullptr) {
		VarjoHMDPlugin = &FModuleManager::LoadModuleChecked<IVarjoHMDPlugin>(TEXT("VarjoHMD"));
	}

	return VarjoHMDPlugin->GetVRSystem();
}

void FVarjoVRController::RegisterDeviceChanges(vr::IVRSystem* VRSystem)
{
	for (uint32 DeviceIndex = 0; DeviceIndex < vr::k_unMaxTrackedDeviceCount; ++DeviceIndex) {
		// see what kind of hardware this is
		vr::ETrackedDeviceClass DeviceClass = VRSystem->GetTrackedDeviceClass(DeviceIndex);

		// TODO: This is a workaround fix for the bug that OpenVR claims VIVE trackers to be controllers
		char* modelName = new char[vr::k_unMaxPropertyStringSize];
		VRSystem->GetStringTrackedDeviceProperty(DeviceIndex, vr::Prop_RenderModelName_String, modelName, vr::k_unMaxPropertyStringSize);
		if (std::string(modelName).find("tracker") != std::string::npos) DeviceClass = vr::TrackedDeviceClass_GenericTracker;
		delete[] modelName;

		switch (DeviceClass)
		{
		case vr::TrackedDeviceClass_Controller:
		{
			// Check connection status
			if (VRSystem->IsTrackedDeviceConnected(DeviceIndex)) {
				// has the controller not been mapped yet
				if (DeviceToControllerMap[DeviceIndex] == INDEX_NONE) {
					RegisterController(DeviceIndex, VRSystem);
				}
			}
			// the controller has been disconnected, unmap it 
			else if (DeviceToControllerMap[DeviceIndex] != INDEX_NONE) {
				UnregisterController(DeviceIndex);
			}
		}
		break;
		case vr::TrackedDeviceClass_GenericTracker:
		{
			// Check connection status
			if (VRSystem->IsTrackedDeviceConnected(DeviceIndex)) {
				// has the tracker not been mapped yet
				if (DeviceToControllerMap[DeviceIndex] == INDEX_NONE) {
					RegisterTracker(DeviceIndex);
				}
			}
			// the tracker has been disconnected, unmap it 
			else if (DeviceToControllerMap[DeviceIndex] != INDEX_NONE) {
				UnregisterTracker(DeviceIndex);
			}
		}
		break;
		case vr::TrackedDeviceClass_Invalid:
			// falls through
		case vr::TrackedDeviceClass_HMD:
			// falls through
		case vr::TrackedDeviceClass_TrackingReference:
			break;
		default:
			UE_LOG(LogVarjoVRController, Warning, TEXT("Encountered unsupported device class of %i!"), (int32)DeviceClass);
			break;
		}
	}
}

bool FVarjoVRController::RegisterController(uint32 DeviceIndex, vr::IVRSystem* VRSystem)
{
	// don't map too many controllers
	if (NumControllersMapped >= MaxControllers) {
		UE_LOG(LogVarjoVRController, Warning, TEXT("Found more controllers than we support (%i vs %i)!  Probably need to fix this."), NumControllersMapped + 1, MaxControllers);
		return false;
	}

	int32 ControllerIndex = FMath::FloorToInt(NumControllersMapped / CONTROLLERS_PER_PLAYER);
	DeviceToControllerMap[DeviceIndex] = ControllerIndex;

	// Decide which hand to associate this controller with
	EControllerHand ChosenHand = EControllerHand::Special_9;
	{
		vr::ETrackedControllerRole Role = VRSystem->GetControllerRoleForTrackedDeviceIndex(DeviceIndex);
		UE_LOG(LogVarjoVRController, Verbose, TEXT("Controller role for device %i is %i (invalid=0, left=1, right=2)."), DeviceIndex, (int32)Role);

		switch (Role)
		{
		case vr::ETrackedControllerRole::TrackedControllerRole_LeftHand:
			ChosenHand = EControllerHand::Left;
			break;
		case vr::ETrackedControllerRole::TrackedControllerRole_RightHand:
			ChosenHand = EControllerHand::Right;
			break;
		case vr::ETrackedControllerRole::TrackedControllerRole_Invalid:
		{
			if (UnrealControllerIdAndHandToDeviceIdMap[DeviceToControllerMap[DeviceIndex]][(int32)EControllerHand::Right] == INDEX_NONE) ChosenHand = EControllerHand::Right;
			else if (UnrealControllerIdAndHandToDeviceIdMap[DeviceToControllerMap[DeviceIndex]][(int32)EControllerHand::Left] == INDEX_NONE) ChosenHand = EControllerHand::Left;
			else return false;
			break;
		}
		default:
			return false;
		}
	}

	ControllerStates[DeviceIndex].Hand = ChosenHand;
	UnrealControllerHandUsageCount[(int32)ChosenHand] += 1;

	++NumControllersMapped;

	UnrealControllerIdAndHandToDeviceIdMap[DeviceToControllerMap[DeviceIndex]][(int32)ControllerStates[DeviceIndex].Hand] = DeviceIndex;

	return true;
}

void FVarjoVRController::DetectHandednessSwap(vr::IVRSystem* VRSystem)
{
	const uint32 LeftDeviceIndex = VRSystem->GetTrackedDeviceIndexForControllerRole(vr::TrackedControllerRole_LeftHand);
	const uint32 RightDeviceIndex = VRSystem->GetTrackedDeviceIndexForControllerRole(vr::TrackedControllerRole_RightHand);

	// both hands need to be assigned
	if (LeftDeviceIndex != vr::k_unTrackedDeviceIndexInvalid && RightDeviceIndex != vr::k_unTrackedDeviceIndexInvalid) {

		// Don't set the handedness for general trackers
		if ((ControllerStates[LeftDeviceIndex].Hand != EControllerHand::Left && ControllerStates[LeftDeviceIndex].Hand != EControllerHand::Right) ||
			(ControllerStates[RightDeviceIndex].Hand != EControllerHand::Left && ControllerStates[RightDeviceIndex].Hand != EControllerHand::Right)) {
			return;
		}

		// see if our mappings don't match
		if (ControllerStates[LeftDeviceIndex].Hand != EControllerHand::Left || ControllerStates[RightDeviceIndex].Hand != EControllerHand::Right) {
			// explicitly assign the handedness
			ControllerStates[LeftDeviceIndex].Hand = EControllerHand::Left;
			ControllerStates[RightDeviceIndex].Hand = EControllerHand::Right;

			int32 ControllerIndex = DeviceToControllerMap[LeftDeviceIndex];

			UnrealControllerIdAndHandToDeviceIdMap[ControllerIndex][(int32)EControllerHand::Left] = LeftDeviceIndex;
			UnrealControllerIdAndHandToDeviceIdMap[ControllerIndex][(int32)EControllerHand::Right] = RightDeviceIndex;
		}
	}
}

bool FVarjoVRController::RegisterTracker(uint32 DeviceIndex)
{
	// check to see if there are any Special designations left, skip mapping it if there are not
	if (NumTrackersMapped >= MaxSpecialDesignations) {
		// go ahead and increment, so we can display a little more info in the log
		++NumTrackersMapped;
		UE_LOG(LogVarjoVRController, Warning, TEXT("Unable to map VR tracker (#%i) to Special hand designation!"), NumTrackersMapped);
		return false;
	}

	// add the tracker to player 0
	DeviceToControllerMap[DeviceIndex] = GENERIC_TRACKER_PLAYER_NUM;

	// select next special #
	switch (NumTrackersMapped)
	{
	case 0:
		ControllerStates[DeviceIndex].Hand = EControllerHand::Special_1;
		break;
	case 1:
		ControllerStates[DeviceIndex].Hand = EControllerHand::Special_2;
		break;
	case 2:
		ControllerStates[DeviceIndex].Hand = EControllerHand::Special_3;
		break;
	case 3:
		ControllerStates[DeviceIndex].Hand = EControllerHand::Special_4;
		break;
	case 4:
		ControllerStates[DeviceIndex].Hand = EControllerHand::Special_5;
		break;
	case 5:
		ControllerStates[DeviceIndex].Hand = EControllerHand::Special_6;
		break;
	case 6:
		ControllerStates[DeviceIndex].Hand = EControllerHand::Special_7;
		break;
	case 7:
		ControllerStates[DeviceIndex].Hand = EControllerHand::Special_8;
		break;
	case 8:
		ControllerStates[DeviceIndex].Hand = EControllerHand::Special_9;
		break;
	case 9:
		ControllerStates[DeviceIndex].Hand = EControllerHand::Special_10;
		break;
	case 10:
		ControllerStates[DeviceIndex].Hand = EControllerHand::Special_11;
		break;
	default:
		// initial mapping verification above should catch any erroneous NumTrackersMapped
		check(false);
		break;
	}

	++NumTrackersMapped;
	UE_LOG(LogVarjoVRController, Log, TEXT("Tracker device %i is being assigned unreal hand: Special %i, for player %i"), DeviceIndex, NumTrackersMapped, GENERIC_TRACKER_PLAYER_NUM);

	UnrealControllerIdAndHandToDeviceIdMap[DeviceToControllerMap[DeviceIndex]][(int32)ControllerStates[DeviceIndex].Hand] = DeviceIndex;

	return true;
}

void FVarjoVRController::UnregisterController(uint32 DeviceIndex)
{
	UnrealControllerHandUsageCount[(int32)ControllerStates[DeviceIndex].Hand] -= 1;
	UnregisterDevice(DeviceIndex);
	NumControllersMapped--;
}

void FVarjoVRController::UnregisterTracker(uint32 DeviceIndex)
{
	UnregisterDevice(DeviceIndex);
	NumTrackersMapped--;
}

void FVarjoVRController::UnregisterDevice(uint32 DeviceIndex)
{
	// undo the mappings
	UnrealControllerIdAndHandToDeviceIdMap[DeviceToControllerMap[DeviceIndex]][(int32)ControllerStates[DeviceIndex].Hand] = INDEX_NONE;
	DeviceToControllerMap[DeviceIndex] = INDEX_NONE;

	// re-zero out the controller state
	FMemory::Memzero(&ControllerStates[DeviceIndex], sizeof(FControllerState));
}

FName FVarjoVRController::DeviceTypeName(TEXT("VarjoVRController"));
EVarjoVRTouchDPadMapping FVarjoVRController::DefaultDPadMapping = EVarjoVRTouchDPadMapping::FaceButtons;

FVarjoVRController* GetVarjoMotionController()
{
	TArray<IMotionController*> MotionControllers = IModularFeatures::Get().GetModularFeatureImplementations<IMotionController>(IMotionController::GetModularFeatureName());
	for (IMotionController* MotionController : MotionControllers) {
		if (MotionController->GetMotionControllerDeviceTypeName() == FVarjoVRController::DeviceTypeName) {
			return static_cast<FVarjoVRController*>(MotionController);
		}
	}
	return nullptr;
}

// defined here in this .cpp file so we have access to FVarjoVRController
void UVarjoVRControllerLibrary::SetTouchDPadMapping(EVarjoVRTouchDPadMapping NewMapping)
{
	// modify the default mapping in case we haven't instantiated a FVarjoVRController yet
	FVarjoVRController::DefaultDPadMapping = NewMapping;

	FVarjoVRController* VarjoMotionController = GetVarjoMotionController();
	if (VarjoMotionController) VarjoMotionController->SetTouchDPadMapping(NewMapping);
}

bool UVarjoVRControllerLibrary::IsControllerConnected(int32 ControllerIndex, FName MotionSource)
{
	bool RetVal = false;
	EControllerHand Hand = EControllerHand::Left;
	FXRMotionControllerBase::GetHandEnumForSourceName(MotionSource, Hand);

	FVarjoVRController* VarjoMotionController = GetVarjoMotionController();
	if (VarjoMotionController) RetVal = VarjoMotionController->IsControllerConnected(ControllerIndex, Hand);

	return RetVal;
}

float UVarjoVRControllerLibrary::GetControllerBatteryLevel(int32 ControllerIndex, FName MotionSource)
{
	float RetVal = 0.0f;
	EControllerHand Hand = EControllerHand::Left;
	FXRMotionControllerBase::GetHandEnumForSourceName(MotionSource, Hand);

	FVarjoVRController* VarjoMotionController = GetVarjoMotionController();
	if (VarjoMotionController) RetVal = VarjoMotionController->GetControllerBatteryLevel(ControllerIndex, Hand);

	return RetVal;
}

class FVarjoVRControllerPlugin : public IVarjoVRControllerPlugin
{
	virtual TSharedPtr< class IInputDevice > CreateInputDevice(const TSharedRef< FGenericApplicationMessageHandler >& InMessageHandler) override
	{
		return TSharedPtr< class IInputDevice >(new FVarjoVRController(InMessageHandler));
	}
};

#undef LOCTEXT_NAMESPACE //"VarjoVRController"

IMPLEMENT_MODULE( FVarjoVRControllerPlugin, VarjoVRController)
