#ifndef VARJO_TYPES_LAYERS_H
#define VARJO_TYPES_LAYERS_H

#include "Varjo_types.h"

#if defined __cplusplus
extern "C" {
#endif

/**
 * Swap chain configuration structure.
 */
struct varjo_SwapChainConfig2 {
    varjo_TextureFormat textureFormat;  //!< Texture format of the swap chain images
    int32_t numberOfTextures;           //!< Number of textures in the swap chain.
    int32_t textureWidth;               //!< Texture width in pixels.
    int32_t textureHeight;              //!< Texture height in pixels.
    int32_t textureArraySize;           //!< Texture array size, or 1 if creating a plain 2D texture swap chain.
};

/** Opaque type for the swap chain object. */
struct varjo_SwapChain;

/** Swap chain image source rectangle.  */
struct varjo_SwapChainViewport {
    struct varjo_SwapChain* swapChain;  //!< The swap chain to retrieve the image from.
    int32_t x;                          //!< X offset within the swap chain for the input rectangle.
    int32_t y;                          //!< Y offset within the swap chain for the input rectangle.
    int32_t width;                      //!< Width of the input rectangle in pixels.
    int32_t height;                     //!< Height of the input rectangle in pixels.
    int32_t arrayIndex;                 //!< Texture array slice (if the swap chain is a texture array), otherwise 0.
    int32_t reserved;                   //!< Reserved for future use.
};

typedef int64_t varjo_LayerType;
/** Layer type enum corresponding to #varjo_LayerMultiProj. */
static const varjo_LayerType varjo_LayerMultiProjType = 0x1;

typedef int64_t varjo_LayerFlags;
/** Default layer rendering */
static const varjo_LayerFlags varjo_LayerFlagNone = 0x0;

static const varjo_LayerFlags varjo_LayerFlag_BlendMode_Inherit = 0x0;     //!< Use blend mode specified in SubmitInfo.
static const varjo_LayerFlags varjo_LayerFlag_BlendMode_Opaque = 0x1;      //!< Override blend mode to be opaque.
static const varjo_LayerFlags varjo_LayerFlag_BlendMode_AlphaBlend = 0x2;  //!< Override blend mode to be alpha blend.
static const varjo_LayerFlags varjo_LayerFlag_DepthSorting = 0x4;          //!< Override depth sorting to enabled .

typedef int64_t varjo_Space;
/**
 * The local space, world-locked origin.
 */
static const varjo_Space varjo_SpaceLocal = 0x0;
/**
 * Reference space relative to the user's head.
 * This space can be used to render face-locked layers.
 */
static const varjo_Space varjo_SpaceView = 0x1;

/**
 * Layer header for different submission layer types.
 */
struct varjo_LayerHeader {
    /** The actual type of the layer; currently, the only supported type is varjo_LayerMultiProjType */
    varjo_LayerType type;
    /** Flags for this layer. See #varjo_LayerFlags. */
    varjo_LayerFlags flags;
};

typedef int64_t varjo_ViewExtensionType;
static const varjo_ViewExtensionType varjo_ViewExtensionDepthType = 0x1;

/**
 * View Extension header.
 */
struct varjo_ViewExtension {
    /** View extension type. Currently the only supported type is #varjo_ViewExtensionDepthType. */
    varjo_ViewExtensionType type;
    /** Optional pointer to the next extension; allows chaining multiple extensions into a single view. */
    struct varjo_ViewExtension* next;
};

/**
 * A #varjo_LayerMultiProjView extension that allows submitting depth surface
 * alongside the color surface in a view.
 */
struct varjo_ViewExtensionDepth {
    /** Extension header, with type = #varjo_ViewExtensionDepthType */
    struct varjo_ViewExtension header;
    /** Smallest depth value allowed in the depth surface, typically 0.0 */
    double minDepth;
    /** Largest depth value allowed in the depth surface, typically 1.0 */
    double maxDepth;
    /** Positive distance in meters. To indicate depth values reversed, can be greater than farZ. */
    double nearZ;
    /** Positive distance in meters. To indicate depth values reversed, can be smaller than nearZ. */
    double farZ;
    /** Image source for the depth surface. */
    struct varjo_SwapChainViewport viewport;
};

/**
 * A single view in a Multi-projection layer
 */
struct varjo_LayerMultiProjView {
    /**
     * An optional extension structure, or nullptr if none.
     * Currently the only supported extension is #varjo_ViewExtensionDepth
     */
    struct varjo_ViewExtension* extension;
    /** The projection matrix this view was rendered with. */
    struct varjo_Matrix projection;
    /** The view matrix this view was rendered with. */
    struct varjo_Matrix view;
    /** Image source for this view. See #varjo_SwapChainViewport for more info. */
    struct varjo_SwapChainViewport viewport;
};

/** Layer type describing a set of planar projected images rendered from the perspective of each eye. */
struct varjo_LayerMultiProj {
    /** The layer header. */
    struct varjo_LayerHeader header;
    /**
     * The reference space for this layer. See #varjo_Space for valid options.
     */
    varjo_Space space;
    /** The amount of entries in the #views array */
    int32_t viewCount;
    /** Array of views to be rendered. See #varjo_LayerMultiProjView for more information. */
    struct varjo_LayerMultiProjView* views;
};

/**
 * A structure describing the list of layers that the application submits to the compositor each frame.
 */
struct varjo_SubmitInfoLayers {
    /**
     * The frame number of the frame being submitted. This should match the frame number returned
     * in #varjo_SubmitInfo when calling #varjo_BeginFrame.
     */
    int64_t frameNumber;
    /**
     * The submission flags, see #varjo_SubmitFlags for a list of valid flags.
     */
    varjo_SubmitFlags flags;
    /** The number of elements in the #layers array below */
    int32_t layerCount;
    /**
     * Array of pointers to layer structures, one for each layer.
     * Note that the elements in this array are not really of #varjo_LayerHeader, but
     * rather structures such as #varjo_LayerMultiProj that contain #varjo_LayerHeader as
     * a first element. #varjo_LayerHeader.type is used to detect the actual structure type.
     * Currently the only supported layer type is #varjo_LayerMultiProjType, that corresponds
     * to #varjo_LayerMultiProj structure.
     *
     * When submitting multiple layers, the layers are rendered back-to-front in the order
     * they're provided in this array.
     */
    struct varjo_LayerHeader** layers;
};

#if defined __cplusplus
}
#endif

#endif
